from threading import Lock
from typing import Dict, Set, List, TYPE_CHECKING, Union, Optional
from dataclasses import dataclass, field
from collections.abc import Iterable
from maspy.utils import manual_deepcopy, merge_dicts

if TYPE_CHECKING:
    from maspy.agent import Agent

DEFAULT_GROUP = "default"

@dataclass
class Percept:
    key: str
    _args: tuple = field(default_factory=tuple)
    group: str = DEFAULT_GROUP # Percept Type (still unsure)
    adds_event: bool = True
    
    @property
    def args(self):
        if len(self._args) > 1:
            return self._args
        elif len(self._args) == 1:
            return self._args[0]
        else:
            return tuple()
    
    @property
    def args_len(self):
        return len(self._args)
    
    def __post_init__(self):
        match self._args:
            case list() | dict() | str():
                object.__setattr__(self, "_args", tuple([self._args]))
            case tuple():
                pass
            case Iterable():
                object.__setattr__(self, "_args", tuple(self._args))
            case _:
                object.__setattr__(self, "_args", tuple([self._args]))
    
    def __hash__(self) -> int:
        args_hashable = []
        for arg in self.args:
            args_hashable.append(arg)

        args_tuple: tuple = tuple(args_hashable)

        return hash((self.key, args_tuple, self.group))

@dataclass
class Model:
    actions: list
    state_type: str = "Discrete"

class EnvironmentMultiton(type):
    _instances: Dict[str, "Environment"] = {}
    _lock: Lock = Lock()

    def __call__(cls, env_name=None,full_log=False):
        with cls._lock:
            _my_name = env_name if env_name else str(cls.__name__)
            if _my_name not in cls._instances:
                vars = []
                if env_name: 
                    vars.append(env_name)
                if full_log: 
                    vars.append(full_log)
                instance = super().__call__(*vars)
                cls._instances[_my_name] = instance
        return cls._instances[_my_name]

class Environment(metaclass=EnvironmentMultiton):
    def __init__(self, env_name: Optional[str]=None,full_log: bool=False):
        self._my_name = env_name if env_name else type(self).__name__
        self.show_exec = full_log
        self.lock = Lock()
        
        from maspy.admin import Admin
        Admin()._add_environment(self)
        #  dict[Agt_cls, dict[Agt_name, set[Agt_fullname]]]
        self.agent_list: dict[str, dict[str, set[tuple]]] = dict()
        #  dict[Agt_fullname, Agt_inst]
        self._agents: dict[tuple, 'Agent'] = dict()
        
        self._name = f"Environment:{self._my_name}"
        #  dict[env_name, dict[percept_key, set[Percept]]]
        self._percepts: dict[str, Dict[str, Set[Percept]]] = dict()
    
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)
    
    def perception(self):
        return manual_deepcopy(self._percepts)

    @property
    def print_percepts(self):
        percepts = ""
        for group_keys in self._percepts.values():
            for percept_set in group_keys.values():
                for percept in percept_set:
                    percepts += f"{percept}\n"
        print(f"{percepts}\r")
        
    def add_agents(self, agents: Union[list['Agent'],'Agent']):
        if isinstance(agents, list):
            for agent in agents:
                self._add_agent(agent)
        else:
            self._add_agent(agents)
    
    def _add_agent(self, agent: 'Agent'):
        assert isinstance(agent.my_name, tuple)
        if type(agent).__name__ in self.agent_list:
            if agent.my_name[0] in self.agent_list[type(agent).__name__]:
                self.agent_list[type(agent).__name__][agent.my_name[0]].update({agent.my_name})
                self._agents[agent.my_name] = agent
            else:
                self.agent_list[type(agent).__name__].update({agent.my_name[0] : {agent.my_name}})
                self._agents[agent.my_name] = agent
        else:
            self.agent_list[type(agent).__name__] = {agent.my_name[0] : {agent.my_name}}
            self._agents[agent.my_name] = agent
        
        self.print(f'Connecting agent {type(agent).__name__}:{agent.my_name}') if self.show_exec else ...
    
    def create(self, percept: list[Percept] | Percept):
        with self.lock:
            percept_dict = self._clean(percept)
            aux_percepts = manual_deepcopy(self._percepts)
            merge_dicts(percept_dict, aux_percepts)
            self._percepts = aux_percepts
            self.print(f"Creating {percept}") if self.show_exec else ...

    def get(self, percept:Percept, all=False, ck_group=False, ck_args=True) -> List[Percept] | Percept | None:
        found_data = []
        for group, keys in self._percepts.items():
            for key, prcs in keys.items():
                for prcpt in prcs:
                    if self._compare_data(prcpt,percept,ck_group,ck_args):
                        if not all:
                            return prcpt
                        else:
                            found_data.append(prcpt)
                            
        return found_data if found_data else None
                    
    def _compare_data(self, data1: Percept, data2: Percept, ck_group:bool,ck_args:bool) -> bool:
        #self.print(f"Comparing: \n\t{data1} and {data2}")
        if ck_group and data1.group != data2.group:
        #    self.print("Failed at group")
            return False
        if data1.key != data2.key:
        #    self.print("Failed at key")
            return False
        if not ck_args:
            return True
        if data1.args_len != data2.args_len:
        #    self.print("Failed at args_len")
            return False
        for arg1,arg2 in zip(data1._args,data2._args):
            if isinstance(arg1,str) and (arg1[0].isupper()):
                continue
            elif isinstance(arg2,str) and (arg2[0].isupper()):
                continue
            elif arg1 == arg2:
                continue
            else:
        #        self.print(f"Failed at args {arg1} x {arg2}")
                return False
        else:
        #    self.print("Data is Compatible")
            return True
    
    def change(self, old_percept:Percept, new_args:Percept.args):
        if type(new_args) is not tuple: 
            new_args = (new_args,) 
        old_percept._args = new_args
            
    def _percept_exists(self, key, args, group=DEFAULT_GROUP) -> bool:
        if type(args) is not tuple: 
            args = (args,)
        return Percept(key,args,group) in self._percepts[group][key]

    def delete(self, percept: list[Percept] | Percept):
        if isinstance(percept, list):
            for prcpt in percept:
                self._percepts[prcpt.group][prcpt.key].remove(prcpt)
        else:
            self._percepts[self._my_name][percept.key].remove(percept)
              
    def _clean(self, percept_data: Iterable[Percept] | Percept) -> Dict:
        match percept_data:
            case None:
                return dict()
            case Percept():
                return {self._my_name : {percept_data.key: {percept_data}}}
            case Iterable():
                percept_dict: dict = dict()
                for prc_dt in percept_data:
                    if prc_dt.key in percept_dict[self._my_name]:
                        percept_dict[self._my_name][prc_dt.key].add(prc_dt)
                    else:
                        percept_dict[self._my_name].update({prc_dt.key: {prc_dt}})

                return percept_dict
            case _:
                raise TypeError(
                    f"Expected data type to have be Iterable[Percept] | Percept, recieved {type(percept_data).__name__}"
                ) 




from maspy.agent import Agent
from maspy.admin import Admin
from maspy.environment import Environment, Percept, DEFAULT_GROUP
from typing import List
import pytest

def test_clean():
    env = Environment("clean")
    assert type(env._clean(None)) is dict
    with pytest.raises(TypeError):
        env._clean(42)
    with pytest.raises(AttributeError):
        env._clean([42])
        
    env._clean([Percept("KEY",(1,)),
                Percept("KEY",(2,)),
                Percept("KEY",(3,)),
                Percept("NEW_KEY")])

def test_percepts_are_Percept():
    env = Environment("test_env1")
    with pytest.raises(Exception):
        env.create()
        
    env.create("KEY")
    env.create("KEY",5)
    env.create("KEY",(1,2,3),"Normal")
    env.create("KEY",group="Normal")
    env.create(percept=Percept("B"))
    env.create(percept=[Percept("C",group="Normal"), Percept("D",(2,))])
    for group, keys in env._percepts.items():
            for key, percept_set in keys.items():
                for percept in percept_set:
                    assert (type(percept) is Percept and percept.key == key and percept.group == group)
                    
def test_delete_percepts():
    env = Environment("test_env2")
    env.create("KEY",23,"Normal")
    env.create("KEY",(42,27))
    assert Percept("KEY",(23,),"Normal") in env._percepts["Normal"]["KEY"]
    assert Percept("KEY",(42,27)) in env._percepts[DEFAULT_GROUP]["KEY"]
    env.delete("KEY",23,"Normal")
    env.delete("KEY",(42,27))
    assert Percept("KEY",(23,),"Normal") not in env._percepts["Normal"]["KEY"]
    assert Percept("KEY",(42,27)) not in env._percepts[DEFAULT_GROUP]["KEY"]
    
def test_change_percept():
    env = Environment("change")
    env.create("KEY",42)
    env.change("KEY",42,27)
    assert Percept("KEY",(42,)) not in env._percepts[DEFAULT_GROUP]["KEY"]
    assert Percept("KEY",(27,)) not in env._percepts[DEFAULT_GROUP]["KEY"]

def test_percept_exists():
    env = Environment("test_env3")
    env.create("KEY",23)
    assert Percept("KEY",(23,)) in env._percepts[DEFAULT_GROUP]["KEY"] and \
        env._percept_exists("KEY",23)
    env.print_percepts

def test_perception():
    env = Environment("test_env1")
    perception_dict, perception_list = env.perception()
    assert perception_dict == env._percepts 
    for percept in perception_list:
        assert (type(percept) is Percept)
        
def test_environment_instances():
    env = Environment()
    env1 = Environment("env1")
    env2 = Environment("env2")
    assert env1 != env2 and env1 == Environment("env1") and env2 == Environment("env2") and env1 != env and env2 != env
    
def test_add_agents():
    ag1 = Agent("Simple")
    ag2 = Agent("Simple")
    ag_list = [Agent(f"Simple{i}") for i in range(10)]
    env = Environment("Adding")
    ag1.connect_to(env)
    ag2.connect_to(env)
    env.add_agents(ag_list)
    ag_list += [ag1,ag2]
    for agent in ag_list:
        assert agent.my_name in env.agent_list["Agent"][agent.my_name[0]]
        assert agent == env._agents[agent.my_name]


def merge_dicts(dict1: dict, dict2: dict) -> dict:
    for key, value in dict1.items():
        if key in dict2 and isinstance(value, dict):
            for inner_key, inner_value in value.items():
                if inner_key in dict2[key] and isinstance(inner_value, set):
                    dict2[key][inner_key].update(inner_value)
                else:
                    dict2[key][inner_key] = inner_value 
        else:
            dict2[key] = value
    return dict2

def set_changes(original:set, changes:set):
    intersection = original.intersection(changes)
    added = changes - original
    removed = original - intersection
    return intersection.union(added), added, removed

def manual_deepcopy(d: dict) -> dict:
    return {k: {k2: set(v2) for k2, v2 in v.items()} for k, v in d.items()}



from threading import Lock
from typing import Any, Dict, List, Union, Optional
from collections.abc import Iterable
from maspy.environment import Environment
from maspy.communication import Channel
from maspy.agent import Agent
import signal
from time import sleep

class AdminMeta(type):
    _instances: Dict[str, Any] = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]
class Admin(metaclass=AdminMeta):
    def __init__(self:'Admin') -> None:
        signal.signal(signal.SIGINT, self.stop_all_agents)
        self.end_of_execution = False
        self._name = f"# {type(self).__name__} #"
        print("Starting MASPY Program")
        self.show_exec = False
        self.agt_sh_exec = False
        self.agt_sh_cycle = False
        self.agt_sh_prct = False
        self.agt_sh_slct = False
        self.ch_sh_exec = False
        self.env_sh_exec = False
        self._started_agents: List[Agent] = list()
        self._agent_list: Dict[tuple, str] = dict()
        self._num_agent: Dict[str, int] = dict()
        self._agents: Dict[tuple, Agent] = dict()
        self._channels: Dict[str, Channel] = dict()
        self._environments: Dict[str, Environment] = dict()
        
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)

    def get_agents(self) -> Dict[tuple, str]:
        return self._agent_list

    def add_agents(
        self, agents: Union[list[Agent], Agent]
    ) -> None:
        if isinstance(agents, list):
            for agent in agents:
                self._add_agent(agent)
        else:
            self._add_agent(agents)

    def _add_agent(self, agent: Agent) -> None:
        name: Optional[str | tuple] = None
        if agent.my_name is None :
            name = type(agent).__name__
        else:
            name = agent.my_name
            
        assert isinstance(name,str)
        if name in self._num_agent:
            self._num_agent[name] += 1
            agent.my_name = (name, self._num_agent[name])
        else:
            self._num_agent[name] = 1
            agent.my_name = (name, 1)
        
        #agent.my_name = ("").join(str(x) for x in agent.name_tag)
        
        self._agent_list[agent.my_name] = type(agent).__name__
        self._agents[agent.my_name] = agent
        agent.full_log = self.agt_sh_exec
        agent.show_cycle = self.agt_sh_cycle
        agent.show_prct = self.agt_sh_prct
        agent.show_slct = self.agt_sh_slct
        self.print(
            f"Registering Agent {type(agent).__name__}:{agent.my_name}"
        ) if self.show_exec else ...

    def rm_agents(
        self, agents: Union[Iterable[Agent], Agent]
    ) -> None:
        if isinstance(agents, list):
            for agent in agents:
                self._rm_agent(agent)
        else:
            assert isinstance(agents, Agent)
            self._rm_agent(agents)

    def _rm_agent(self, agent: Agent):
        if agent.my_name in self._agents:
            assert isinstance(agent.my_name, tuple)
            del self._agents[agent.my_name]
            del self._agent_list[agent.my_name]
        self.print(
            f"Removing agent {type(agent).__name__}:{agent.my_name} from List"
        ) if self.show_exec else ...

    def _add_channel(self, channel: Channel) -> None:
        self._channels[channel._my_name] = channel
        channel.show_exec = self.ch_sh_exec
        self.print(
            f"Registering {type(channel).__name__}:{channel._my_name}"
        ) if self.show_exec else ...

    def _add_environment(self, environment: Environment) -> None:
        self._environments[environment._my_name] = environment
        environment.show_exec = self.env_sh_exec
        self.print(
            f"Registering Environment {type(environment).__name__}:{environment._my_name}"
        ) if self.show_exec else ...
    
    def start_system(self:'Admin') -> None:
        no_agents = True
        try:
            self.print("Starting Agents")
            for agent_name in self._agents:
                no_agents = False
                self._start_agent(agent_name)
            
            if no_agents:
                self.print("No agents are connected")
            
            sleep(1)
            while not self.end_of_execution and self.running_agents():
                sleep(1)
        except Exception:
            pass
        print("Ending MASPY Program")
        
    def running_agents(self:'Admin') -> bool:
        for agent in self._agents.values():
            if agent.running:
                return True
        return False

    def start_agents(
        self, agents: Union[list[Agent], Agent]
    ) -> None:
        if isinstance(agents, list):
            self.print("Starting listed agents")
            for agent in agents:
                assert isinstance(agent.my_name, tuple)
                self._start_agent(agent.my_name)
        else:
            assert isinstance(agents, Agent)
            self.print(f"Starting agent {type(agents).__name__}:{agents.my_name}")
            assert isinstance(agents.my_name, tuple)
            self._start_agent(agents.my_name)

    def _start_agent(self, agent_name: tuple) -> None:
        try:
            if agent_name in self._started_agents:
                self.print(f"'Agent' {agent_name} already started")
                return

            agent = self._agents[agent_name]
            self._started_agents.append(agent)
            agent.reasoning()
        except KeyError:
            self.print(f"'Agent' {agent_name} not connected to environment")
            
    def stop_all_agents(self,sig,frame):
        self.print("[Closing System]")
        for agent in self._agents.values():
            agent.stop_cycle()
        self.end_of_execution = True
    
    def connect_to(self, agents: Iterable | Agent, targets: Iterable[Environment | Channel] | Environment | Channel) -> None:
        if not isinstance(agents, Iterable): 
            agents = [agents]
        if not isinstance(targets, Iterable): 
            targets = [targets]
        for agent in agents:
            for target in targets:
                match target:
                    case Environment():
                        agent._environments[target._my_name] = target
                    case Channel():
                        agent._channels[target._my_name] = target
                
                target._add_agent(agent)

    def set_logging(self, show_exec: bool, show_cycle: bool=False,
                    show_prct: bool=False, show_slct: bool=False, 
                     set_admin=True,
                     set_agents=True,
                     set_channels=True,
                     set_environments=True
                    ) -> None:
        self.show_exec = True if show_exec and set_admin else False
        self.agt_sh_exec = True if show_exec and set_agents else False
        self.agt_sh_cycle = True if show_cycle and set_agents else False
        self.agt_sh_prct = True if show_prct and set_agents else False
        self.agt_sh_slct  = True if show_slct and set_agents else False
        self.ch_sh_exec = True if show_exec and set_channels else False
        self.env_sh_exec = True if show_exec and set_environments else False

    def slow_cycle_by(self, time: int | float) -> None:
        for agent in self._agents.values():
            agent.sleep = time



from .agent import *
from .communication import *
from .environment import *
from .admin import Admin

__all__ = ['Agent','Belief','Goal','Ask','Plan','Event','gain','lose','test','pl',
           'Channel','tell','untell','tellHow','untellHow','achieve','unachieve','askOne', 'askAll','askHow','broadcast',
           'Environment','Percept',
           'Admin']



from setuptools import setup, find_packages

VERSION = '0.2.0'
DESCRIPTION = 'Multi-Agent Systems in Python with Machine Learning'
LONG_DESCRIPTION = 'A library for the devolopment of multi-agent systems with components of machine learning'

# Setting up
setup(
    name="maspy-ml",
    version=VERSION,
    author="Alexandre Mellado",
    author_email="<melladoallm@gamil.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'autonomous agents', 'multi-agent system', 'machine learning'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
    ]
)


from maspy.agent import *
from maspy.communication import Channel
from maspy.admin import Admin
import pytest

def test_plan():
    pass

def test_add_rm_agent():
    ag1 = Agent("ag1")
    ag2 = Agent("ag2")
    Channel("Special").add_agents([ag1,ag2])
    ag1.send(ag2.my_name,tell,("KEY",),"Special")
    Channel("Special")._rm_agents([ag1,ag2])
    Channel()._rm_agents(ag2)
    ag1.send(ag2.my_name,tell,("KEY",27))

def test_send_msg():
    ag1 = Agent("ag1",instant_mail=True)
    ag2 = Agent("ag2",instant_mail=True)
    ag1.send(ag2.my_name,"sending",("KEY",))
    assert ag2.has_close(Belief("KEY")) == False
    ag1.send(ag2.my_name,tell,("KEY",))
    assert ag2.has_close(Belief("KEY")) == True
    ag1.send(ag2.my_name,tellHow,test_plan)
    ag1.send(ag2.my_name,askOne,("ASK",))
    ag1.send(ag2.my_name,untell,("KEY",))
    ag1.send(ag2.my_name,achieve,("OBJ",))
    ag1.send(ag2.my_name,unachieve,("OBJ",))
    ag1.send(ag2.my_name,tell,("KEY",),"VIP_Channel")

    
 


from threading import Lock
from collections.abc import Iterable
from typing import Dict, TypeVar, TYPE_CHECKING, Union, Any

if TYPE_CHECKING:
    from maspy.agent import Agent, Belief, Goal, Ask, Plan

Agt_name = TypeVar('Agt_name', bound=str)
Agt_cls = TypeVar('Agt_cls', bound=str)
Agt_inst = TypeVar('Agt_inst', bound=object)
Agt_fullname = TypeVar('Agt_fullname', bound=tuple[str,int])

tell= TypeVar('tell')
untell = TypeVar('untell')
tellHow = TypeVar('tellHow')
untellHow = TypeVar('untellHow')
achieve = TypeVar('achieve')
unachieve = TypeVar('unachieve')
askOne = TypeVar('askOne')
askAll = TypeVar('askAll')
askHow = TypeVar('askHow')

ACT = Union[tell,untell,tellHow,untellHow,achieve,unachieve,askOne,askAll,askHow]

broadcast = TypeVar('broadcast')


def is_broadcast(target: Any) -> bool:
    return target == broadcast

class CommsMultiton(type):
    _instances: Dict[str, "Channel"] = {}
    _lock: Lock = Lock()

    def __call__(cls, __my_name="default"):
        with cls._lock:
            if __my_name not in cls._instances:
                instance = super().__call__(__my_name)
                cls._instances[__my_name] = instance
        return cls._instances[__my_name]


class Channel(metaclass=CommsMultiton):
    def __init__(self, comm_name:str="default"):
        self.show_exec = False
        
        from maspy.admin import Admin
        self._my_name = comm_name
        Admin()._add_channel(self)
        
        from maspy.agent import Belief, Goal, Ask, Plan
        self.data_types = {Belief,Goal,Ask,Plan}
        self._my_name = comm_name
        # dict[Agt_cls, dict[Agt_name, set[Agt_fullname]]]
        self.agent_list: dict[str, dict[str, set[tuple]]] = dict()
        # dict[Agt_fullname, Agt_inst]
        self._agents: dict[tuple, 'Agent'] = dict()
        self._name = f"{type(self).__name__}:{self._my_name}"
        
    def print(self,*args, **kwargs):
        return print(f"{self._name}>",*args,**kwargs)
    
    def add_agents(self, agents: Union[list['Agent'],'Agent']):
        if isinstance(agents, list):
            for agent in agents:
                self._add_agent(agent)
        else:
            self._add_agent(agents)

    def _add_agent(self, agent: 'Agent'):
        assert isinstance(agent.my_name, tuple)
        if type(agent).__name__ in self.agent_list:
            if agent.my_name[0] in self.agent_list[type(agent).__name__]:
                self.agent_list[type(agent).__name__][agent.my_name[0]].update({agent.my_name})
                self._agents[agent.my_name] = agent
            else:
                self.agent_list[type(agent).__name__].update({agent.my_name[0] : {agent.my_name}})
                self._agents[agent.my_name] = agent
        else:
            self.agent_list[type(agent).__name__] = {agent.my_name[0] : {agent.my_name}}
            self._agents[agent.my_name] = agent
        
        self.print(f'Connecting agent {type(agent).__name__}:{agent.my_name}') if self.show_exec else ...

            
    def _rm_agents(self, agents: Union[list['Agent'],'Agent']):
        if isinstance(agents, list):
            for agent in agents:
                self._rm_agent(agent)
        else:
            self._rm_agent(agents)

    def _rm_agent(self, agent: 'Agent'):
        assert isinstance(agent.my_name, tuple)
        if agent.my_name in self._agents:
            del self._agents[agent.my_name]
            del self.agent_list[type(agent).__name__][agent.my_name[0]]
        self.print(
            f"Desconnecting agent {type(agent).__name__}:{agent.my_name}"
        ) if self.show_exec else ...

    def _send(self, sender: tuple, target: Union[Agt_name, Iterable[Agt_name], broadcast], act: ACT | str, message: Union[str, 'Belief', 'Goal', 'Ask', 'Plan'] | list[Union[str, 'Belief', 'Goal', 'Plan']]):  
        if type(act) == TypeVar: 
            act = act.__name__ 
        
        messages = []
        if isinstance(message, list):
            for m in message:
                messages.append(self.parse_sent_msg(sender,act,m))
        else:
            messages.append(self.parse_sent_msg(sender,act,message))

        for msg in messages:
            if type(target) == Iterable:
                for trgt in target:
                    self._sending(sender,trgt,act,msg)
            elif is_broadcast(target):
                for agent_name in self._agents.keys():
                    if agent_name != sender:
                        self._sending(sender,agent_name,act,msg)
            else:
                self._sending(sender,target,act,msg)
    
    def _sending(self, sender, target, act, msg):
        self.print(f"{sender} sending {act}:{msg} to {target}") if self.show_exec else ...        
    
        try:         
            self._agents[target].save_msg(act,msg)
        except KeyError:
            self.print(f"Agent {target} not connected")
    
    def parse_sent_msg(self,sender, act, msg):
        from maspy.agent import Belief, Goal, Ask
        if type(msg) in {Belief, Goal} and msg is not None:
            msg = msg.update(source=sender)
        if act in {"askOne","askAll","askHow"}:
            msg = Ask(msg, source=sender)
        return msg